from urllib.parse import urlsplit

from celery.result import AsyncResult
from django.contrib.auth.models import User
from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.utils.translation import gettext_lazy as _


class DataLocation(models.Model):
    name = models.CharField(max_length=50, primary_key=True)
    uri = models.CharField(max_length=200)

    @staticmethod
    def insert_location_from_string(location_string):
        """
        Insert a datalocation from a srm string (e.g. srm://srm-url:4321)
        Data Location names are: Sara, Juelich and Poznan
        :param str location_string: SRM url
        :return: DataLocation object
        """
        # netloc will become srm-url:4321
        _, netloc, *_ = urlsplit(location_string)
        if "sara" in netloc.lower():
            loc_name = "Sara"
        elif "fz-juelich" in netloc.lower():
            loc_name = "Juelich"
        elif "psnc" in netloc.lower():
            loc_name = "Poznan"
        else:
            loc_name = "Unknown"

        dataloc = DataLocation(name=loc_name, uri=netloc)
        dataloc.save()
        return dataloc


class DataProduct(models.Model):
    obs_id = models.CharField(verbose_name="OBS_ID", max_length=15)
    oid_source = models.CharField(verbose_name="OBS_ID_SOURCE", max_length=15)
    dataproduct_source = models.CharField(max_length=20)
    dataproduct_type = models.CharField(max_length=50)
    project = models.CharField(max_length=50)
    location = models.ForeignKey(DataLocation, on_delete=models.DO_NOTHING)
    activity = models.CharField(max_length=50)
    surl = models.CharField(max_length=200)
    filesize = models.PositiveBigIntegerField()
    additional_meta = models.JSONField()

    @staticmethod
    def insert_dataproduct(
            obs_id,
            oid_source,
            dataproduct_source,
            dataproduct_type,
            project,
            activity,
            surl,
            filesize,
            additional_meta,
    ):
        scheme, netloc, *_ = urlsplit(surl)

        dp = DataProduct(
            obs_id=obs_id,
            oid_source=oid_source,
            dataproduct_source=dataproduct_source,
            dataproduct_type=dataproduct_type,
            project=project,
            location=DataLocation.insert_location_from_string(
                "://".join((scheme, netloc))
            ),
            activity=activity,
            filesize=filesize,
            additional_meta=additional_meta,
            surl=surl,
        )
        dp.save()
        return dp


class DataFilterType(models.TextChoices):
    DROPDOWN = 'Dropdown', _('Dropdown')
    FREEFORM = 'Free', _('Freeform')


class DataProductFilter(models.Model):
    field = models.CharField(max_length=100)
    name = models.CharField(max_length=20)
    lookup_type = models.CharField(max_length=100)

    filter_type = models.CharField(max_length=20, choices=DataFilterType.choices, default=DataFilterType.FREEFORM)

    def __str__(self):
        return f'({self.pk}) "{self.name}" [{self.field} ->{self.lookup_type}]'


class ATDBProcessingSite(models.Model):
    name = models.CharField(primary_key=True, max_length=100)
    url = models.URLField()
    access_token = models.CharField(max_length=1000, null=True)

    def __str__(self):
        return str(self.name) + ' - ' + str(self.url)


class SUBMISSION_STATUS(models.TextChoices):
    """Status of Work Specifcation to ATDB"""

    NOT_SUBMITTED = "N", _("not submitted")
    # Tasks are submitted, check ATDB for current status
    SUBMITTED = "S", _("submitted")
    # Task need to be submitted (can change predecessor etc...)
    DEFINING = "D", _("defining")
    # Check log, retry submission
    ERROR = "E", _("error")


class PURGE_POLICY(models.TextChoices):
    """ATDB Purge Policy"""

    # @Mancini, what are these?
    YES = "yes", _("yes")
    NO = "no", _("no")
    DO = "do", _("do")  # only purge, no workflow execution


class WorkSpecification(models.Model):
    """Work Specification"""

    created_on = models.DateTimeField(auto_now_add=True)
    created_by = models.ForeignKey(User, on_delete=models.DO_NOTHING, null=True)
    filters = models.JSONField(null=True)
    # Input data containing sizes/urls for submission to ATDB
    inputs = models.JSONField(null=True)

    # ATDB Workflow URL
    selected_workflow = models.CharField(max_length=500, null=True)

    # Task ID's that were created in ATDB
    related_tasks = ArrayField(models.IntegerField(), null=True)
    predecessor_task = models.IntegerField(null=True)
    predecessor_specification = models.ForeignKey('self', null=True, on_delete=models.DO_NOTHING,
                                                  related_name='successor')

    # The query for gathering files has been executed
    is_ready = models.BooleanField(default=False)
    is_defined = models.BooleanField(default=False)

    # Should automatically submit the task after defining
    is_auto_submit = models.BooleanField(default=False)

    async_task_result = models.CharField(max_length=100, null=True)
    processing_site = models.ForeignKey(
        ATDBProcessingSite, null=True, on_delete=models.DO_NOTHING
    )
    purge_policy = models.CharField(
        max_length=16, choices=PURGE_POLICY.choices, default=PURGE_POLICY.NO
    )

    def __str__(self):
        return str(self.id) + ' - ' + str(self.filters) + " (" + str(self.created_on) + ")"

    # How many files per task. 0 is single task with all files
    batch_size = models.IntegerField(default=0, null=False, blank=False)
    submission_status = models.CharField(
        max_length=1,
        choices=SUBMISSION_STATUS.choices,
        default=SUBMISSION_STATUS.NOT_SUBMITTED,
    )

    def save(
            self, force_insert=False, force_update=False, using=None, update_fields=None
    ):
        super(WorkSpecification, self).save(
            force_insert=force_insert,
            force_update=force_update,
            using=using,
            update_fields=update_fields,
        )
        if self.async_task_result is None:
            from lofardata.tasks import define_work_specification

            res: AsyncResult = define_work_specification.delay(self.pk)
            self.async_task_result = res.id
            super(WorkSpecification, self).save(update_fields=["async_task_result"])
