import re
from unittest import mock

import rest_framework.test as rtest
from lofardata.models import SUBMISSION_STATUS, ATDBProcessingSite, WorkSpecification
from lofardata.tasks import SessionStore, delete_tasks_from_atdb
from lofardata.tests.util import MockResponse, mocked_delay

task_detail_uri = re.compile(r"^.*/atdb/tasks/(?P<task_id>\d+)/")
# This is needed to initialize a session
SessionStore.get_session()

# ATDB Post Requests
def mocked_requests_delete(*args, **kwargs):

    url = args[0]

    # Broken task
    if "500" in url:
        return MockResponse(None, 500, "Server Error.")

    if task_detail_uri.match(url):
        # TODO: add real ATDB response
        return MockResponse(None, 204)

    return MockResponse(None, 404)


@mock.patch(
    "lofardata.tasks.SessionStore._session.delete", side_effect=mocked_requests_delete
)
class TestATDBDeleteRequest(rtest.APITestCase):
    @mock.patch(
        "lofardata.tasks.define_work_specification.delay", side_effect=mocked_delay
    )
    def setUp(self, mock_delay: mock.MagicMock):
        site: ATDBProcessingSite = ATDBProcessingSite.objects.create(
            name="Dwingeloo", url="https://example.com/atdb/", access_token="DummyToken"
        )

        # Create Work Specification
        ws: WorkSpecification = WorkSpecification.objects.create(processing_site=site)
        mock_delay.assert_called_once_with(ws.pk)  # not really needed here actually
        ws.refresh_from_db()

        # Simulated interaction where the workflow was set
        ws.selected_workflow = "https://example.com/atdb/workflow/1/"
        ws.save()
        self.work_spec: WorkSpecification = ws

    def test_delete(self, mocked_delete: mock.MagicMock):
        self.work_spec.related_tasks = [1]  # just a single task
        self.work_spec.save()

        delete_tasks_from_atdb(self.work_spec.pk)

        mocked_delete.assert_called_once_with(
            f"https://example.com/atdb/tasks/{self.work_spec.related_tasks[0]}/",
            auth=mock.ANY,
        )

        self.work_spec.refresh_from_db()
        self.assertEqual(
            self.work_spec.submission_status,
            SUBMISSION_STATUS.NOT_SUBMITTED,
            "Status should be not submitted after delete",
        )
        self.assertIsNone(
            self.work_spec.related_tasks,
            "Related tasks should be deleted after delete request",
        )

    def test_delete_of_batches(self, mocked_delete: mock.MagicMock):
        self.work_spec.related_tasks = [1, 2, 3, 4]
        self.work_spec.save()

        delete_tasks_from_atdb(self.work_spec.pk)

        call_list = [
            mock.call(f"https://example.com/atdb/tasks/{task_id}/", auth=mock.ANY)
            for task_id in self.work_spec.related_tasks
        ]

        mocked_delete.assert_has_calls(call_list, any_order=True)

        self.work_spec.refresh_from_db()
        self.assertEqual(
            self.work_spec.submission_status,
            SUBMISSION_STATUS.NOT_SUBMITTED,
            "Status should be not submitted after delete",
        )
        self.assertIsNone(
            self.work_spec.related_tasks,
            "All related tasks should be deleted after delete request",
        )

    def test_delete_of_batches_with_failing(self, _: mock.MagicMock):
        self.work_spec.related_tasks = [1, 2, 500, 4]
        self.work_spec.save()

        delete_tasks_from_atdb(self.work_spec.pk)

        self.work_spec.refresh_from_db()
        self.assertEqual(
            self.work_spec.submission_status,
            SUBMISSION_STATUS.ERROR,
            "Status should be error after delete failed",
        )
