import re
from unittest import mock

import rest_framework.test as rtest
from lofardata.models import SUBMISSION_STATUS, ATDBProcessingSite, WorkSpecification
from lofardata.tasks import SessionStore, set_tasks_defined
from lofardata.tests.util import MockResponse, mocked_delay

task_detail_uri = re.compile(r"^.*/atdb/tasks/(?P<task_id>\d+)/")
# This is needed to initialize a session
SessionStore.get_session()

# ATDB Post Requests
def mocked_requests_put(*args, **kwargs):

    url = args[0]

    # Broken task
    if "500" in url:
        return MockResponse(None, 500, "Server Error.")

    if task_detail_uri.match(url):
        # TODO: add real ATDB response
        return MockResponse({"id": 42}, 200)

    return MockResponse(None, 404)


@mock.patch(
    "lofardata.tasks.SessionStore._session.put", side_effect=mocked_requests_put
)
class TestATDBTaskStatusChange(rtest.APITestCase):
    @mock.patch(
        "lofardata.tasks.define_work_specification.delay", side_effect=mocked_delay
    )
    def setUp(self, mock_delay: mock.MagicMock):
        site: ATDBProcessingSite = ATDBProcessingSite.objects.create(
            name="Dwingeloo", url="https://example.com/atdb/", access_token="DummyToken"
        )

        # Create Work Specification
        ws: WorkSpecification = WorkSpecification.objects.create(processing_site=site)
        mock_delay.assert_called_once_with(ws.pk)  # not really needed here actually
        ws.refresh_from_db()

        # Simulated interaction where the workflow was set
        ws.selected_workflow = "https://example.com/atdb/workflow/1/"
        ws.save()
        self.work_spec: WorkSpecification = ws

    def test_request_submission(self, mocked_put: mock.MagicMock):

        self.work_spec.related_tasks = [1]  # just a single task
        self.work_spec.submission_status = SUBMISSION_STATUS.DEFINING
        self.work_spec.save()

        expected_payload = {"new_status": "defined"}

        set_tasks_defined(self.work_spec.pk)

        mocked_put.assert_called_once_with(
            f"https://example.com/atdb/tasks/{self.work_spec.related_tasks[0]}/",
            json=expected_payload,
            auth=mock.ANY,
        )

        self.work_spec.refresh_from_db()
        self.assertAlmostEqual(
            self.work_spec.submission_status, SUBMISSION_STATUS.SUBMITTED
        )

    def test_request_submission_of_batches(self, mocked_put: mock.MagicMock):

        self.work_spec.related_tasks = [1, 2, 3]
        self.work_spec.submission_status = SUBMISSION_STATUS.DEFINING
        self.work_spec.save()

        expected_payload = {"new_status": "defined"}

        set_tasks_defined(self.work_spec.pk)

        call_list = [
            mock.call(
                f"https://example.com/atdb/tasks/{task_id}/",
                json=expected_payload,
                auth=mock.ANY,
            )
            for task_id in self.work_spec.related_tasks
        ]

        mocked_put.assert_has_calls(call_list, any_order=True)

        self.work_spec.refresh_from_db()
        self.assertAlmostEqual(
            self.work_spec.submission_status, SUBMISSION_STATUS.SUBMITTED
        )

    def test_request_submission_of_batches_with_failing_task(self, _: mock.MagicMock):

        self.work_spec.related_tasks = [1, 2, 500, 4]
        self.work_spec.submission_status = SUBMISSION_STATUS.DEFINING
        self.work_spec.save()

        set_tasks_defined(self.work_spec.pk)

        self.work_spec.refresh_from_db()
        self.assertEquals(
            self.work_spec.submission_status,
            SUBMISSION_STATUS.ERROR,
            "Status should be error after put failure",
        )
