from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from django.shortcuts import redirect
from django.shortcuts import render
from django.urls import reverse
from django_filters import rest_framework as filters
from rest_framework import generics, status, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.schemas.openapi import AutoSchema

from .forms import WorkSpecificationForm
from .models import DataProduct, DataProductFilter, DataLocation, WorkSpecification, ATDBProcessingSite, DataFilterType
from .serializers import DataProductSerializer, \
    DataProductFlatSerializer, DataLocationSerializer, \
    WorkSpecificationSerializer, ATDBProcessingSiteSerializer
from .tasks import insert_task_into_atdb


class DynamicFilterSet(filters.FilterSet):
    class Meta:
        filter_class = None

    def __init__(self, *args, **kwargs):
        self._load_filters()
        super().__init__(*args, **kwargs)

    def _load_filters(self):
        if self.Meta.filter_class is None:
            raise Exception('Define filter_class meta attribute')
        for item in self.Meta.filter_class.objects.all():
            field_obj = self.Meta.model._meta.get_field(item.field)
            filter_class, *_ = self.filter_for_lookup(field_obj, item.lookup_type)
            self.base_filters[item.name] = filter_class(item.field)


# --- Filters ---
class DataProductFilterSet(DynamicFilterSet):
    class Meta:
        model = DataProduct
        filter_class = DataProductFilter
        fields = {
            "obs_id": ["exact", "icontains"],
        }


# ---------- GUI Views -----------


def index(request):
    atdb_hosts = ATDBProcessingSite.objects.values("name", "url")
    return render(request, "lofardata/index.html", {"atdb_hosts": atdb_hosts})


def api(request):
    atdb_hosts = ATDBProcessingSite.objects.values('name', 'url')
    return render(request, "lofardata/api.html", {'atdb_hosts': atdb_hosts})


def construct_specifications_list():
    results = ''
    for specification in WorkSpecification.objects.all():
        url = reverse('specification-detail', args=(str(specification.pk),))
        line = f'<li><a class="dropdown-item" href="{url}">{str(specification)}</a></li>'
        results += line
    return results


def handle_specification_view_post(request, specification):
    form = WorkSpecificationForm(request.POST, instance=specification)
    if form.is_valid():
        object_to_save = request.POST['save']

        if object_to_save == 'save':
            specification.async_task_result = None
            specification.is_ready = False
            specification.save()
            return redirect('specification-detail', specification.pk)
        elif object_to_save == 'save_and_send_to_atdb':
            specification.save()
            res = insert_task_into_atdb.delay(specification.pk)
            return redirect('specification-detail', specification.pk)
        elif object_to_save == 'save_and_add_successor':
            specification.save()

            successor = WorkSpecification()
            successor.predecessor_specification = specification
            successor.processing_site = specification.processing_site
            successor.save()
            return redirect('specification-detail', successor.pk)


def preprocess_filters_specification_view(specification):
    dataproduct_filters = DataProductFilter.objects.all()
    for dataproduct_filter in dataproduct_filters:
        if specification.filters and dataproduct_filter.field in specification.filters:
            dataproduct_filter.default = specification.filters[dataproduct_filter.field]
        else:
            dataproduct_filter.default = ''

        if dataproduct_filter.filter_type == DataFilterType.DROPDOWN:
            dataproduct_filter.choices = DataProduct.objects.distinct(dataproduct_filter.field).values_list(
                dataproduct_filter.field)
    return dataproduct_filters


def specification_view(request, pk=None):
    try:
        # retrieve existing specification
        specification = WorkSpecification.objects.get(pk=pk)
    except ObjectDoesNotExist:
        # start new specification
        specification = WorkSpecification()

    specifications_list = construct_specifications_list()
    # a POST means that the form is filled in and should be stored in the
    form = WorkSpecificationForm(instance=specification)
    if request.method == "POST":
        return handle_specification_view_post(request, specification)

    dataproduct_filters = preprocess_filters_specification_view(specification)

    return render(request, "lofardata/specification.html", {
        'form': form,
        'specification': specification,
        'specifications_list': specifications_list,
        'filters': dataproduct_filters
    })


# ---------- REST API views ----------
class DataProductView(generics.ListCreateAPIView):
    model = DataProduct
    serializer_class = DataProductSerializer

    queryset = DataProduct.objects.all().order_by("obs_id")

    # using the Django Filter Backend - https://django-filter.readthedocs.io/en/latest/index.html
    filter_backends = (filters.DjangoFilterBackend,)
    filter_class = DataProductFilterSet


class ATDBProcessingSiteView(viewsets.ReadOnlyModelViewSet):
    model = ATDBProcessingSite
    serializer_class = ATDBProcessingSiteSerializer

    queryset = ATDBProcessingSite.objects.all().order_by('pk')


class DataProductDetailsView(generics.RetrieveUpdateDestroyAPIView):
    model = DataProduct
    serializer_class = DataProductSerializer
    queryset = DataProduct.objects.all()


class DataLocationView(generics.ListCreateAPIView):
    model = DataLocation
    serializer_class = DataLocationSerializer
    queryset = DataLocation.objects.all().order_by("name")


class InsertWorkSpecificationSchema(AutoSchema):
    def get_operation_id_base(self, path, method, action):
        return "createDataProductMulti"


class InsertMultiDataproductView(generics.CreateAPIView):
    """
    Add single DataProduct
    """

    queryset = DataProduct.objects.all()
    serializer_class = DataProductFlatSerializer
    schema = InsertWorkSpecificationSchema()

    def get_serializer(self, *args, **kwargs):
        """if an array is passed, set serializer to many"""
        if isinstance(kwargs.get("data", {}), list):
            kwargs["many"] = True
        return super().get_serializer(*args, **kwargs)


class WorkSpecificationViewset(viewsets.ModelViewSet):
    queryset = WorkSpecification.objects.all()
    serializer_class = WorkSpecificationSerializer

    def get_queryset(self):
        current_user: User = self.request.user
        if not current_user.is_staff or not current_user.is_superuser:
            return self.queryset.filter(created_by=current_user.id)
        else:
            return self.queryset

    @action(detail=True, methods=["POST"])
    def submit(self, request, pk=None) -> Response:
        # TODO: check that there are some matches in the request?
        # TODO: how to specify the filter?
        res = insert_task_into_atdb.delay(pk)
        return Response({"detail": "accepted"}, status=status.HTTP_202_ACCEPTED)
